#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import re
import random
from sysrepo_interfaces import vpp as vpputils

ovpn_client_file = "/config/etc/openvpn/client.conf"

def start_openvpn(self):
    self.execute_shell_command("systemctl enable openvpn@client.service")
    if not os.path.isfile('/config/.hamode'):
        try:
            self.execute_shell_command("systemctl --no-block start openvpn@client.service")
        except:
            self.logger.warning("Can not start openvpn")

def stop_openvpn(self):
    try:
        self.execute_shell_command("systemctl disable openvpn@client.service")
        self.execute_shell_command("systemctl --no-block stop openvpn@client.service")
        self.execute_shell_command("rm -f /etc/network/interfaces.d/ifcfg-tun0")
    except:
        self.logger.warning("Can not stop openvpn")

def restart_openvpn(self):
    do_restart = True

    if os.path.isfile('/config/.hamode'):
        do_restart = False
        try:
            if str(self.execute_shell_command("systemctl is-active openvpn@client.service")).strip("\n") == "active":
                do_restart = True
        except:
            pass

    if not do_restart:
        return

    try:
        self.execute_shell_command("systemctl stop openvpn@client.service")
        self.execute_shell_command("systemctl --no-block start openvpn@client.service")
    except:
        self.logger.warning("Can not restart openvpn")

def config_openvpn_ippass(self, old_port, new_port, old_enable, new_enable, startup):
    self.logger.debug('config_openvpn_ippass({},{},{},{}, {})'.format(old_port, new_port, old_enable, new_enable, startup))
    if startup and new_enable:
        self.logger.debug('OVPN startup config - {} {}'.format(new_enable, new_port))
        vpputils.set_ippass_enable_udp_dst_port(self, new_port, new_enable)

    if not new_enable and old_enable:
        # OVPN have to be disabled
        # remove entries for old port
        self.logger.debug('OVPN is disabled - remove entries for port {}'.format(old_port))
        vpputils.set_ippass_enable_udp_dst_port(self, old_port, new_enable)

    if not old_enable and new_enable:
        # OVPN is enabled
        # add entries for new_port
        self.logger.debug('OVPN is enabled - add entries for port {}'.format(new_port))
        vpputils.set_ippass_enable_udp_dst_port(self, new_port, new_enable)

    if old_enable and new_enable:
        if new_port != old_port:
            # port is changed
            self.logger.debug('OVPN is enabled - remove entries for port {}'.format(old_port))
            vpputils.set_ippass_enable_udp_dst_port(self, old_port, False)
            self.logger.debug('OVPN is enabled - add entries for port {}'.format(new_port))
            vpputils.set_ippass_enable_udp_dst_port(self, new_port, True)

def config_openvpn_interface(self, name, config, change_set):
    global ovpn_client_file

    if not config:
        self.logger.debug('{} no config'.format(name))
        return

    self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))
    enable_vpn = False
    if 'enable' in config:
        enable_vpn = config['enable']

    old_enable_vpn = enable_vpn

    concentrator_endpoint = None
    concentrator_port = None
    if 'ovpn-cfg' in config:
        if 'concentrator-endpoint' in config['ovpn-cfg']:
            concentrator_endpoint = config['ovpn-cfg']['concentrator-endpoint']
        if 'concentrator-port' in config['ovpn-cfg']:
            concentrator_port = config['ovpn-cfg']['concentrator-port']

    old_port = concentrator_port
    old_enable = old_enable_vpn

    if change_set:
        for xpath, change in change_set.items():
            if change['operation'] == 'DELETED':
                continue
            if xpath.find('enable') > -1:
                enable_vpn = change['new_value']
            if xpath.find('concentrator-endpoint') > -1:
                concentrator_endpoint = change['new_value']
            if xpath.find('concentrator-port') > -1:
                concentrator_port = change['new_value']
                
    #get ports from current config of openvpn
    try:
        with open (ovpn_client_file, "rt") as openvpnf:
            vpncfg=openvpnf.read()
            matches=re.search("\nremote (.*) (.*)\n",vpncfg)
            if matches:
                curr_conf_conc=(matches.group(1)).strip()
                curr_conf_port=(matches.group(2)).strip()
                if old_port:
                    old_port = curr_conf_port
    except:
        curr_conf_conc=None
        curr_conf_port=None
        
    #concentrator port can be a list
    if concentrator_port:
        concentrator_portslist = sum([ list(range(int(x.split("-")[0]),int(x.split("-")[1])+1)) for x in concentrator_port.split(",") if "-" in x],
                                     [ int(x) for x in concentrator_port.split(",") if "-" not in x  ])
        concentrator_port = random.choice(concentrator_portslist)
        port_in_list =  [ele for ele in concentrator_portslist if(str(ele) == str(curr_conf_port))]
        if bool(port_in_list):
            concentrator_port = curr_conf_port

    new_port = concentrator_port
    new_enable = enable_vpn
    config_openvpn_ippass(self, old_port, new_port, old_enable, new_enable, False if change_set else True)

    if concentrator_port and concentrator_endpoint:
        if concentrator_endpoint.startswith("tcp6:"):
            concentrator_endpoint=concentrator_endpoint.replace("tcp6:","")
            self.execute_shell_command("sed -i 's/^proto .*/proto tcp6-client/' %s" % ovpn_client_file)
            self.execute_shell_command("sed -i 's/^fragment .*/#fragment 1300/' %s" % ovpn_client_file)
        elif concentrator_endpoint.startswith("tcp:"):
            concentrator_endpoint=concentrator_endpoint.replace("tcp:","")
            self.execute_shell_command("sed -i 's/^proto .*/proto tcp-client/' %s" % ovpn_client_file)
            self.execute_shell_command("sed -i 's/^fragment .*/#fragment 1300/' %s" % ovpn_client_file)
        elif concentrator_endpoint.startswith("udp6:"):
            concentrator_endpoint=concentrator_endpoint.replace("udp6:","")
            self.execute_shell_command("sed -i 's/^proto .*/proto udp6/' %s" % ovpn_client_file)
            self.execute_shell_command("sed -i 's/^#fragment .*/fragment 1300/' %s" % ovpn_client_file)
        else:
            concentrator_endpoint=concentrator_endpoint.replace("udp:","")
            concentrator_endpoint=concentrator_endpoint.replace("udp4:","")
            self.execute_shell_command("sed -i 's/^proto .*/proto udp/' %s" % ovpn_client_file)
            self.execute_shell_command("sed -i 's/^#fragment .*/fragment 1300/' %s" % ovpn_client_file)

        self.execute_shell_command("sed -i 's/^remote.*/remote\ %s\ %s/' %s" % (
            concentrator_endpoint, concentrator_port, ovpn_client_file))
        self.execute_shell_command("sed -i 's/^lport.*/lport\ %s/' %s" % (
            concentrator_port, ovpn_client_file))

    if 'true' in str(old_enable_vpn).lower() and 'true' in str(enable_vpn).lower():
        restart_openvpn(self)
    else:
        if 'true' in str(enable_vpn).lower():
            start_openvpn(self)
        else:
            stop_openvpn(self)
