#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import re
import subprocess

from sysrepo_interfaces import utils as utils
from sysrepo_interfaces import vpp as vpputils
from sysrepo_interfaces import ipsecvpn as ipsec
import vppapi_if.vppapi_if as vppif
import pyroute2 as prt
from tveif import if_lte
from sysrepo_interfaces import vpp as vpputils

USE_TAP=True
BR_PREFFIX='br-'
VETH_PREFFIX='v'
ENABLE_RT_CTL=False
ENABLE_DRCTL=True

if_mac = dict()

def init_mac_db(intf_cfg):
    for ifname, ifcfg in intf_cfg.items():
        mac = None
        if if_lte.lte_check_raw_if(ifname):
            mac = if_lte.lte_get_raw_mac(ifname)
        else:
            if os.path.exists('/sys/class/net/' + ifname + '/address'):
                f = open('/sys/class/net/' + ifname + '/address', 'r')
                mac = f.readline()[0:17]
            elif os.path.exists('/sys/class/net/br-' + ifname + '/address'):
                f = open('/sys/class/net/br-' + ifname + '/address', 'r')
                mac = f.readline()[0:17]

        if mac:
            if_mac[ifname] = mac

def create_bridge(self, vpp_if, ifname, ifid, ifmode=None):
    #Create br-XXX and attach to bridge-domain -- set mac ???
    working_mac=if_mac.get(ifname.split('.')[0],None)
    if USE_TAP:
        (retval, bridx)=vpp_if.create_tap_interface('{}{}'.format(BR_PREFFIX,ifname), host_mac_addr=working_mac)
    else:
        with prt.IPRoute() as ipr:
            # create interface pair
            ipr.link('add', ifname='{}{}'.format(BR_PREFFIX,ifname), kind='veth', peer='{}{}'.format(VETH_PREFFIX,ifname))
            index = ipr.link_lookup(ifname='{}{}'.format(BR_PREFFIX,ifname))[0]
            ipr.link('set', index=index, address=working_mac)
            (retval, bridx)=vpp_if.create_host_interface('{}{}'.format(VETH_PREFFIX,ifname))

    self.logger.debug('Created interface {}{} with ID {}'.format(BR_PREFFIX,ifname,bridx))
    if 0 == retval:
        self.logger.debug('Add interface ID {} to bridge-domain {}'.format(bridx,ifid))
        vpp_if.add_to_bridge(ifid,bridx)

    if ifmode == 'virtio':
        subprocess.check_call(['/usr/sbin/ip', 'link', 'set', ifname, 'promisc', 'on'])
        #try to set tx queue len in virtio mode
        os.system("/usr/sbin/ifconfig {}{} txqueuelen 512 || true".format(BR_PREFFIX,ifname))
    os.system('for offl in `ethtool -k %s%s | grep -v fixed | grep ": on" | cut -f1 -d":" | xargs` ; do ethtool -K %s%s $offl off  ; done' % (BR_PREFFIX,ifname,BR_PREFFIX,ifname) )
    #Up interface
    vpp_if.ifup(bridx)

    if ENABLE_RT_CTL:
        #Set Loopback BVI
        self.logger.info('Loop interface')
        (retval, loop_idx) = vpp_if.create_loopback_interface(mac_addr=working_mac)
        self.logger.info('Loop interface create {} {}'.format(retval, loop_idx))
        vpp_if.add_to_bridge(ifid,loop_idx, port_type=vppif.L2_API_PORT_TYPE_BVI)
        self.logger.info('Loop interface 2')
        vpp_if.set_interface_learning(bridx, disable=1)
        vpp_if.enable_routectl(bridx, loop_idx)
        vpp_if.routectl_set_mac(working_mac)
        vpp_if.ifup(loop_idx)

    if ENABLE_DRCTL:
        (retval, loop_idx) = vpp_if.create_loopback_interface(mac_addr=working_mac)
        self.logger.info('Loop interface create {} {}'.format(retval, loop_idx))
        vpp_if.add_to_bridge(ifid,loop_idx, port_type=vppif.L2_API_PORT_TYPE_BVI)
        vpp_if.set_interface_learning(bridx, disable=1)
        vpp_if.enable_drctl(bridx, mac=working_mac)
        vpp_if.drctl_set_arp(bridx, loop_idx)
        vpp_if.drctl_set_dhcp(bridx, loop_idx)
        vpp_if.ifup(loop_idx)


def create_vpp_vlan(self, ifname, change_set):
    # set_vpp_vlan creates sub-interface adds it to a  newly created bridge-domain
    # returns it's index
    if self.get_network_mode() != 'vpp':
        return

    if ifname is None:
        return

    if change_set:
        vpp_if = vppif.VppApiInterface('sr-interfaces')

        base_name, vlan = str(ifname).split('.')
        ifid = vpp_if.get_ifidx_by_name(str(vpputils.get_vpp_ifname_from_mode(self, ifname)).split('.')[0])

        if ifid is None:
            self.logger.error('Failed to create sub-interface {}{}, because base interface \'host-/{}\' was not found'.format(BR_PREFFIX, ifname, base_name))
            del vpp_if
            return
        if vlan is not None and vlan != 0:
            self.logger.debug('Created sub-interface to ifid {} with vlan {}'.format(ifid, int(vlan)))
            (retval, subidx) = vpp_if.create_vlan_subinterface(idx=ifid, vlan_id=int(vlan))
            self.logger.debug('Created sub-interface {}{} with ID {} vlan {}'.format(BR_PREFFIX, ifname, subidx, int(vlan)))
            if retval == 0 and vpp_if.vlan_rewrite_poptag1(subidx) == 0:
                ret = vpp_if.create_bridge_domain(subidx, tag='{}{}'.format(BR_PREFFIX,ifname))
                self.logger.debug('Create Bridge \'{}\' returns {}'.format(subidx, ret))
                vpp_if.add_to_bridge(subidx, subidx)
                vpp_if.ifup(subidx)
                ifid = subidx

        create_bridge(self, vpp_if, ifname, ifid)
        del vpp_if


def delete_vpp_vlan(self, config, ifname):
    if self.get_network_mode() != 'vpp':
        return
    vpp_if = vppif.VppApiInterface('sr-interfaces')

    base_name, vlan = str(ifname).split('.')
    base_name = base_name.split('-')[-1]
    ifid = None
    bd_id = vpp_if.get_bd_id_by_tag(ifname)
    bd = vpp_if.get_bridge_domain(bd_id)
    if bd and 'sw_if_details' in bd:
        for iface in bd['sw_if_details']:
            ret = vpp_if.del_from_bridge(bd_id, iface['sw_if_index'])
            self.logger.debug("Deleting interface {} from bridge domain {} returned {}".format(iface['sw_if_index'], bd['bd_id'], ret))
            iface_name = vpp_if.get_if_by_ifidx(iface['sw_if_index'])['ifname']
            if vpp_if.is_vlan_sub_interface(iface_name):
                ret = vpp_if.delete_vlan_subinterface(iface['sw_if_index'])
            elif vpp_if.is_tap_interface(iface_name):
                ret = vpp_if.delete_tap_interface(iface['sw_if_index'])
            else:
                ret = vpp_if.delete_loopback_interface(iface['sw_if_index'])
            self.logger.debug("Deleting interface {} returned {}".format(iface['sw_if_index'], bd['bd_id'], ret))

    ret = vpp_if.delete_bridge_domain(bd_id)
    self.logger.debug("Deleting bridge domain with ID {} returned {}".format(bd_id, ret))

    # Delete bridge-domains which are empty or ones with single irrelevant sub-interface
    bridge_domains = vpp_if.get_bridge_domains()
    for bd in bridge_domains:
        if bd['n_sw_ifs'] == 0:
            ret = vpp_if.delete_bridge_domain(bd['bd_id'])
            self.logger.debug("Deleting bridge domain with ID {} returned {}".format(bd['bd_id'], ret))
        if bd['n_sw_ifs'] == 1:
            if_details = bd['sw_if_details']
            ifname = vpp_if.get_ifname_by_ifidx(if_details[0]['sw_if_index'])
            if vpp_if.is_vlan_sub_interface(ifname):
                ret = vpp_if.delete_vlan_subinterface(vpp_if.get_ifidx_by_name(ifname))
                ret = vpp_if.delete_bridge_domain(bd['bd_id'])
                self.logger.debug("Deleting bridge domain with ID {} returned {}".format(bd['bd_id'], ret))
    del vpp_if


def config_interface_vlan(self, name, config, change_set, startup=False):
    self.logger.debug("config_interface_vlan")
    self.logger.debug("Config {}".format(config))
    self.logger.debug("Change set {}".format(change_set))

    enable = ""

    if_names = [f for f in os.listdir('/sys/class/net') if
                os.path.isfile(os.path.join('/sys/class/net', f)) or
                os.path.islink(os.path.join('/sys/class/net', f))]

    basename = re.sub(r'(.*)(\.[0-9]+)$', r'\1', name)  # vlan tag interface - check for base name
    if basename not in if_names \
       and 'dpdk' != self.get_interface_param(basename, 'mode', 'virtio'):
            self.logger.info("Interface {} not found in the system! Skipping configuration...".format(name))
            return

    bootproto = self.get_interface_param(name, 'bootproto', 'dhcp')
    ipv4_prefix = self.get_interface_param(name, 'ipv4-prefix', '')
    ipv6_prefix = self.get_interface_param(name, 'ipv6-prefix', '')
    dns = self.get_interface_param(name, 'dns', '')
    dns6 = self.get_interface_param(name, 'dns6', '')
    gateway = self.get_interface_param(name, 'gateway', '')
    gateway6 = self.get_interface_param(name, 'gateway6', '')
    metric = self.get_interface_param(name, 'metric', 100)
    management = self.get_interface_param(name, 'management', '')
    management_old = 'false'

    ifname = utils.get_ifname(self, name)

    if change_set:
        for xpath, change in change_set.items():
            if change['operation'] == 'DELETED':
                # Fill defaults
                if xpath.find('bootproto') > -1:
                    bootproto = 'dhcp'
                if xpath.find('ipv4-prefix') > -1:
                    ipv4_prefix = ''
                if xpath.find('ipv6-prefix') > -1:
                    ipv6_prefix = ''
                if xpath.find('dns6') > -1:
                    dns6 = ''
                elif xpath.find('dns') > -1:
                    dns = ''
                if xpath.find('metric') > -1:
                    metric = 100
                if xpath.find('gateway6') > -1:
                    gateway6 = ''
                elif xpath.find('gateway') > -1:
                    gateway = ''
                if xpath.find('management') > -1:
                    management = 'false'
                if xpath.find('enable') > -1 and 'smart-protocol-passthrough' not in xpath:
                    enable = 'false'
                if xpath.find('name') > -1:
                    enable = 'false'
                continue

            if xpath.find('enable') > -1 and "tracehost" not in xpath and 'smart-protocol-passthrough' not in xpath:
                enable = change['new_value']
            if xpath.find('bootproto') > -1:
                bootproto = change['new_value']
            if xpath.find('ipv4-prefix') > -1:
                ipv4_prefix = change['new_value']
            if xpath.find('ipv6-prefix') > -1:
                ipv6_prefix = change['new_value']
            if xpath.find('dns6') > -1:
                dns6 = change['new_value']
            elif xpath.find('dns') > -1:
                dns = change['new_value']
            if xpath.find('metric') > -1:
                metric = change['new_value']
            if xpath.find('gateway6') > -1:
                gateway6 = change['new_value']
            elif xpath.find('gateway') > -1:
                gateway = change['new_value']
            if xpath.find('management') > -1:
                management = change['new_value']
                if 'old_value' in change:
                    management_old = change['old_value']

    if str(enable).lower() == 'false':
        self.logger.info("Disabling VLAN iface")
        utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(ifname))
        self.execute_shell_command("/scripts/apply_networking.py -z {}".format(ifname))
        utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname))
        utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(name))
        self.execute_shell_command("/scripts/apply_networking.py -z {}".format(name))
        utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(name))
        delete_vpp_vlan(self, config, ifname)
        return
    elif str(enable).lower() == 'true':
        if not startup:
             create_vpp_vlan(self, name, change_set)

    if bootproto not in ['dhcp', 'dhcp6', 'dhcp46', 'manual', 'static']:
        bootproto = None

    if not bootproto:
        if ipv4_prefix or ipv6_prefix:
            bootproto = "static"
        else:
            bootproto = "manual"

    onbt = "yes"

    self.logger.info('{} onbt is {}'.format(name, onbt))

    utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(ifname))
    utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(ifname))

    utils.shell_return_ok(self,"ifdown --ignore-error --force {}".format(name))
    utils.shell_return_ok(self,"rm -rf /etc/network/interfaces.d/ifcfg-{}".format(name))
#    options = ["-n {}".format(name), "-o {}".format('yes'), "-b manual", "-p \'sysctl -w net.ipv6.conf.{}.disable_ipv6=1\'".format(name)]
    if bootproto.find("6") < 0 or not ipv6_prefix:
        utils.shell_return_ok(self,"echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(name))
#    self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))
    utils.do_ifup(self, name)

    utils.apply_management_rules(self, ifname, management_old, management, change_set)

    options = ["-n {}".format(ifname), "-o {}".format(onbt)]
    if bootproto:
        options.append("-b {}".format(bootproto))
    if ipv4_prefix:
        options.append("--ipv4-prefix={}".format(ipv4_prefix))
    if ipv6_prefix:
        options.append("--ipv6-prefix={}".format(ipv6_prefix))
        options.append('-p"echo 0 > /proc/sys/net/ipv6/conf/%s/accept_dad"' % str(ifname))
    if dns:
        options.append("-d {}".format(dns))
    if dns6:
        options.append("--ipv6-dns={}".format(dns6))
    if gateway:
        options.append("-g {}".format(gateway))
    if gateway6:
        options.append("--ipv6-gateway={}".format(gateway6))
        #options.append('-j"ip -6 route add {gateway} dev {interface} && ip -6 route add default via {gateway}"'.format(gateway=gateway6, interface=ifname))
        options.append('-k"ip -6 route del {gateway} dev {interface} && ip -6 route del default via {gateway}"'.format(gateway=gateway6, interface=ifname))
    if metric:
        options.append("-c {}".format(metric))
    if ifname.find("br") > -1:
        options.append("--ovs-master ovs --ovs-ports {}".format(name))
        if bootproto.find("6") < 0 or not ipv6_prefix:
            utils.shell_return_ok(self,"echo 0 > /proc/sys/net/ipv6/conf/{}/autoconf".format(ifname))

    mac_address=if_mac.get(name.split('.')[0],None)
    if mac_address is not None:
        options.append(f'--hwaddress {mac_address}')
    self.execute_shell_command("/scripts/apply_networking.py {}".format(" ".join(options)))

    try: # onbt == "yes"
        utils.do_ifup(self, ifname)
    except subprocess.CalledProcessError as ex:
        self.logger.error('ifup command failed!')

    if dns:
        dnses = dns.split(',')
        dnsopts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dnses))            
        subprocess.call('/usr/bin/resolvectl -4 dns {} {}'.format(name, dnsopts), shell=True)
    if dns6:
        dns6es = dns6.split(',')
        dns6opts = ''.join(map(lambda dns_srv: ' '+ dns_srv, dns6es))            
        subprocess.call('/usr/bin/resolvectl -6 dns {} {}'.format(ifname, dns6opts), shell=True)

    vpputils.config_interface_ippass(self, name, config, change_set)

    # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Disconnect first.
    ipsecvpn_reenable_list = ipsec.ipsecvpn_for_reenable(self, ifname)
    ipsecvpn_reenable_count = len(ipsecvpn_reenable_list)
    if ipsecvpn_reenable_count > 0:
        self.logger.debug("{} ipsecvpn(s) need to be reenabled!".format(ipsecvpn_reenable_count))
        for ipsec_iface in ipsecvpn_reenable_list:
            ipsec.ipsecvpn_disconnect(self, ipsec_iface)

    try:
        utils.restore_mac(name, ifname, ifmode)
    except:
        pass

    if ipv4_prefix or ipv6_prefix:
        vpputils.handle_vpp_interface_addresses(self, ifname, ipv4_prefix, ipv6_prefix)
        if self.get_kubernetes_bind_interface() == name:
            # We prioritize IPv6
            ip_prefix = ipv6_prefix or ipv4_prefix
            ip = ip_prefix.split('/')[0]
            #some legacy code left here, lets notify it also
            subprocess.Popen(['/scripts/set_openstack_address.sh', f'{ip}'])
            subprocess.Popen(['/scripts/handle_k8s_ip_change.sh', f'{ip}'])

    utils.config_smart_protocol_passthrough(self, name, config, change_set)

    # Reconnect all ipsecvpn that have local IP equal to IP of adaptor that is going to be down. Connect.
    for ipsec_iface in ipsecvpn_reenable_list:
        ipsec.ipsecvpn_connect(self, ipsec_iface)

    ipsecvpn_reenable_list = []
